/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.extension;

import static org.fortiss.af3.component.utils.ComponentArchitectureUtils.isAtomicComponent;
import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.createDataDictionary;
import static org.fortiss.af3.generator.common.utils.CLanguageModelElementFactory.createCSourcePackage;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getConfigureFile;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getEclipseCDTProjectFile;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getEclipseProjectFile;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getMakedefsFile;
import static org.fortiss.af3.platform.utils.PlatformArchitectureUtils.findAtomicTransmissionUnits;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getRootElement;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.tooling.kernel.utils.TransformationUtils.createTransformedObjectFor;
import static org.fortiss.tooling.kernel.utils.TransformationUtils.createTransformedObjectWithoutExceptionFor;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.conqat.lib.commons.collections.Pair;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.component.model.generator.ComponentProgram;
import org.fortiss.af3.deployment.generator.DeploymentExecutable;
import org.fortiss.af3.deployment.generator.DeploymentExecutableTransformationContext;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.expression.model.DataDictionary;
import org.fortiss.af3.expression.model.terms.IExpressionTerm;
import org.fortiss.af3.generator.common.model.c.AF3GeneratorCommonLanguagesCFactory;
import org.fortiss.af3.generator.common.model.c.CSourcePackage;
import org.fortiss.af3.platform.language.executable.ExecutableBase;
import org.fortiss.af3.platform.language.executable.ExecutionUnitExecutableBase;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.af3.platform.raspberry.AF3PlatformRaspberryActivator;
import org.fortiss.af3.platform.raspberry.generator.framework.IRasPiHeaderGeneratorExtension;
import org.fortiss.af3.platform.raspberry.generator.framework.IRasPiLibraryBasedGeneratorExtension;
import org.fortiss.af3.platform.raspberry.generator.framework.IRasPiSourceBasedGeneratorExtension;
import org.fortiss.af3.platform.raspberry.model.CanBus;
import org.fortiss.af3.platform.raspberry.model.RaspberryPi;
import org.fortiss.tooling.kernel.extension.data.ITransformationContext;
import org.fortiss.tooling.kernel.extension.exception.ChainTransformationFailedException;

/** {@link ExecutionUnitExecutableBase} for the {@link RaspberryPi} platform. */
public class RaspberryPIGeneratorExtension extends ExecutionUnitExecutableBase<RaspberryPi, CSourcePackage> {
	/** Folder name for library source files. */
	private static final String SRC_LIB_SUB_PACKAGE_NAME = "src-lib";
	/** Folder name for library include files. */
	public static final String INC_LIB_SUB_PACKAGE_NAME = "inc-lib";
	/** Folder name for linkable static library files. */
	private static final String LIB_SUB_PACKAGE_NAME = "lib";
	/** Default library names to be used always. */
	// Note that the order is important (GCC linker crap!)
	private static final String LIB_NAMES = "af3pihal rt pthread";
	/** Default library names extended with brick. */
	// Note that the order is important (GCC linker crap!)
	private static final String LIB_NAMES_WITH_BRICK = "brick " + LIB_NAMES;
	/** Default library names extended with vesc. */
	// Note that the order is important (GCC linker crap!)
	private static final String LIB_NAMES_WITH_VESC = LIB_NAMES_WITH_BRICK + " vesc";

	// code generator input and intermediary data
	/** The transformation context provided by the deployment generator framework. */
	private ITransformationContext context;
	/** The current deployment. */
	private Deployment deployment;
	/** If the deployment uses a {@link CanBus} this references its catalog. */
	private CanTransmissionCatalog canCatalog;
	/** The list of atomic components deployed on this execution unit. */
	private List<Component> atomics = new ArrayList<>();
	/** The map from platform elements to executables. */
	private Map<PlatformConnectorUnit, ExecutableBase<?>> platformConnector2ExecutableBase =
			new HashMap<>();

	// code generator output
	/** The C source package produced by this generator. */
	private CSourcePackage generatorResult;

	/** Constructor. */
	public RaspberryPIGeneratorExtension(RaspberryPi modelElement) {
		super(modelElement);
	}

	/** {@inheritDoc} */
	@Override
	protected CSourcePackage createExecutionUnitSourcePackage(String name,
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, ITransformationContext context) {
		this.context = context;
		findDeploymentAndLookForCanTransmissionCatalog();
		createGeneratorResultSourcePackage(name);
		try {
			addEclipseCProjectFiles();
			addConfigureAndMakedefsFiles();
			Set<String> headers = addPlatformElementsCode(deployedPorts);
			addDataDictionaryCode();
			addLogicalComponentCode(deployedComponents);
			addMainFile(deployedComponents, deployedPorts, headers);
		} catch(Exception ex) {
			error(AF3PlatformRaspberryActivator.getDefault(), ex.getMessage(), ex);
			ex.printStackTrace();
		}
		return generatorResult;
	}

	/**
	 * Adds the libraries and header files required by the used platform elements and returns the
	 * set of header file names.
	 */
	private Set<String> addPlatformElementsCode(
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts) {
		Set<String> includedHeaders = new HashSet<>();
		Set<String> includedSources = new HashSet<>();
		Set<String> includedLibraries = new HashSet<>();
		for(Pair<PlatformConnectorUnit, Port> pair : deployedPorts) {
			PlatformConnectorUnit pcu = pair.getFirst();
			ExecutableBase<?> executable =
					createTransformedObjectWithoutExceptionFor(pcu, ExecutableBase.class, context);
			platformConnector2ExecutableBase.put(pcu, executable);
			if(executable instanceof IRasPiHeaderGeneratorExtension) {
				IRasPiHeaderGeneratorExtension headerExec = (IRasPiHeaderGeneratorExtension)executable;
				if(!includedHeaders.contains(headerExec.getHeaderFileName())) {
					generatorResult.getSubPackage(INC_LIB_SUB_PACKAGE_NAME).addUnit(
							headerExec.getHeaderFileContent());
					includedHeaders.add(headerExec.getHeaderFileName());
				}
				// ignore header file, since it is already included
			}
			if(executable instanceof IRasPiSourceBasedGeneratorExtension) {
				IRasPiSourceBasedGeneratorExtension srcExec = (IRasPiSourceBasedGeneratorExtension)executable;
				if(!includedSources.contains(srcExec.getSourceFileName())) {
					generatorResult.getSubPackage(SRC_LIB_SUB_PACKAGE_NAME).addUnit(
							srcExec.getSourceFileContent());
					includedSources.add(srcExec.getSourceFileName());
				}
				// ignore source file, since it is already included
			}
			if(executable instanceof IRasPiLibraryBasedGeneratorExtension) {
				IRasPiLibraryBasedGeneratorExtension libExec = (IRasPiLibraryBasedGeneratorExtension)executable;
				if(!includedLibraries.contains(libExec.getLibraryFileName())) {
					generatorResult.getSubPackage(LIB_SUB_PACKAGE_NAME).addUnit(
							libExec.getLibraryFileContent());
					includedLibraries.add(libExec.getLibraryFileName());
				}
				// ignore library file, since it is already included
			}
		}
		return includedHeaders;
	}

	/** Initializes the generator result source package. */
	private void createGeneratorResultSourcePackage(String name) {
		// create the result package
		generatorResult = createCSourcePackage();
		generatorResult.setBaseLocation(name);
		CSourcePackage incLib =
				AF3GeneratorCommonLanguagesCFactory.eINSTANCE.createCSourcePackage();
		incLib.setBaseLocation(INC_LIB_SUB_PACKAGE_NAME);
		generatorResult.getSubPackages().add(incLib);

		CSourcePackage lib = AF3GeneratorCommonLanguagesCFactory.eINSTANCE.createCSourcePackage();
		lib.setBaseLocation(LIB_SUB_PACKAGE_NAME);
		generatorResult.getSubPackages().add(lib);

		CSourcePackage srcLib =
				AF3GeneratorCommonLanguagesCFactory.eINSTANCE.createCSourcePackage();
		srcLib.setBaseLocation(SRC_LIB_SUB_PACKAGE_NAME);
		generatorResult.getSubPackages().add(srcLib);
	}

	/** Adds the code for the deployed logical components. */
	private void addLogicalComponentCode(List<Pair<ExecutionUnit, Component>> deployedComponents)
			throws ChainTransformationFailedException {
		if(canCatalog != null) {
			findAtomicDeployedComponents(deployedComponents);
			addLogicalComponentCodeMultiUnit();
		} else {
			addLogicalComponentCodeSingleUnit(deployedComponents);
		}
	}

	/** Searches the deployed components for atomic ones and adds them to the {@code atomics} list. */
	private void findAtomicDeployedComponents(
			List<Pair<ExecutionUnit, Component>> deployedComponents) {
		atomics.clear();
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			if(p.getFirst() != modelElement) {
				continue;
			}
			Component c = p.getSecond();
			if(isAtomicComponent(c)) {
				atomics.add(c);
			} else {
				atomics.addAll(c.findAtomicComponents());
			}
		}
	}

	/**
	 * Analyzes the deployment context searching for {@link Deployment} and
	 * {@link CanTransmissionCatalog}.
	 */
	private void findDeploymentAndLookForCanTransmissionCatalog() {
		deployment = null;
		canCatalog = null;
		if(context instanceof DeploymentExecutableTransformationContext) {
			DeploymentExecutableTransformationContext dContext =
					(DeploymentExecutableTransformationContext)context;
			DeploymentExecutable dExec = dContext.getDeploymentExecutable();
			deployment = dExec.getDeployment();
			List<TransmissionUnit> tus =
					findAtomicTransmissionUnits(modelElement.getPlatformArchitecture());
			for(TransmissionUnit tu : tus) {
				if(tu instanceof CanBus) {
					canCatalog = (CanTransmissionCatalog)dExec.getTransmissionCatalog(tu);
					break;
				}
			}
		}
	}

	/** Creates the main file by using the {@link SingleUnitMainGenerator} helper class. */
	private void addMainFile(List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, Set<String> headers) {
		CSourcePackage srcGenPack = (CSourcePackage)generatorResult.getSrcGenPackage();
		if(canCatalog != null) {
			MultiUnitMainGenerator mg =
					new MultiUnitMainGenerator(deployment, canCatalog, modelElement, atomics,
							deployedComponents, deployedPorts, context);
			srcGenPack.addUnit(mg.createMultiUnitMain());
		} else {
			SingleUnitMainGenerator mg =
					new SingleUnitMainGenerator(modelElement, deployedComponents, deployedPorts,
							platformConnector2ExecutableBase, headers, context);
			srcGenPack.addUnit(mg.createSingleUnitMain());
		}
	}

	/** Adds the build process files: configure and Makedefs. */
	private void addConfigureAndMakedefsFiles() {
		generatorResult.addUnit(getConfigureFile(modelElement.getName(), LIB_NAMES_WITH_VESC));
		generatorResult.addUnit(getMakedefsFile());
	}

	/** Adds auxiliary files for Eclipse C project. */
	private void addEclipseCProjectFiles() {
		String name = modelElement.getName();
		generatorResult.addUnit(getEclipseProjectFile(name));
		generatorResult.addUnit(getEclipseCDTProjectFile(name));
	}

	/** Creates the code for all deployed components. */
	private void addLogicalComponentCodeMultiUnit() throws ChainTransformationFailedException {
		for(Component cmp : atomics) {
			ComponentProgram cprog =
					createTransformedObjectFor(cmp, ComponentProgram.class, context);
			if(cprog != null) {
				CSourcePackage cPack =
						createTransformedObjectFor(cprog, CSourcePackage.class, context);
				if(cPack != null) {
					cPack.mergeInto(generatorResult);
				}
			}
		}
	}

	/** Creates the code for all deployed components. */
	private void addLogicalComponentCodeSingleUnit(
			List<Pair<ExecutionUnit, Component>> deployedComponents)
			throws ChainTransformationFailedException {
		for(Pair<ExecutionUnit, Component> pair : deployedComponents) {
			if(pair.getFirst() != modelElement) {
				continue;
			}
			Component cmp = pair.getSecond();
			ComponentProgram cprog =
					createTransformedObjectFor(cmp, ComponentProgram.class, context);
			if(cprog != null) {
				CSourcePackage cPack =
						createTransformedObjectFor(cprog, CSourcePackage.class, context);
				if(cPack != null) {
					cPack.mergeInto(generatorResult);
				}
			}
		}
	}

	/** Creates the code for the data dictionary. */
	private void addDataDictionaryCode() throws ChainTransformationFailedException {
		DataDictionary dd = getRootElement(modelElement, DataDictionary.class);
		if(dd == null) {
			dd = createDataDictionary();
		}
		CSourcePackage ddPackage = createTransformedObjectFor(dd, CSourcePackage.class, context);
		if(dd != null) {
			ddPackage.mergeInto(generatorResult);
		}
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getInitialization() {
		// not used
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getTermination() {
		// not used
		return null;
	}
}
